@@ -1,21 +1,53 @@
 package Apache::Htpasswd;
 
+# $Id: Htpasswd.pm,v 1.5.1 2001/03/26 08:25:38 kevin Exp kevin $
+# $Log: Htpasswd.pm,v $
+
+# Revision 1.5.1  2001/03/26 08:25:38 kevin
+# Fixed another newline problem
+
+# Revision 1.5  2001/03/15 01:50:12 kevin
+# Fixed bug to remove newlines
+
+# Revision 1.4  2001/02/23 08:23:46 kevin
+# Added support for extra info fields
+
+# Revision 1.3  2000/04/04 15:00:15 meltzek
+# Made file locking safer to avoid race conditions. Fixed
+# typo in docs.  
+
+# Revision 1.2  1999/01/28 22:43:45  meltzek
+# Added slightly more verbose error croaks. Made sure error from htCheckPassword is only called when called directly, and not by $self.
+#
+# Revision 1.1  1998/10/22 03:12:08  meltzek
+# Slightly changed how files lock.
+# Made more use out of carp and croak.
+# Made sure there were no ^M's as per Randal Schwartz's request.
+#
+# Revision 1.0  1998/10/21 05:53:56  meltzek
+# First version on CPAN.
+#
+
+
 use vars qw(@ISA @EXPORT @EXPORT_OK %EXPORT_TAGS $VERSION);
 use strict;		# Restrict unsafe variables, references, barewords
-#use Carp;
+use Carp;
+
+use POSIX qw ( SEEK_SET SEEK_END );
+use Fcntl qw ( LOCK_EX LOCK_UN );
 
 @ISA = qw(Exporter);
 
 @EXPORT = qw();
 
-@EXPORT_OK = qw(htpasswd htDelete fetchPass htCheckPassword error Version);
+@EXPORT_OK = qw(htpasswd htDelete fetchPass fetchInfo writeInfo htCheckPassword error Version);
 
 %EXPORT_TAGS = (all => [@EXPORT_OK]);
 
-$VERSION = '0.9';
+($VERSION = substr(q$Revision: 1.5.1 $, 10)) =~ s/\s+$//;
 
 sub Version {
-return $VERSION;
+	return $VERSION;
 }
 
 #-----------------------------------------------------------#
@@ -23,13 +55,17 @@ return $VERSION;
 #-----------------------------------------------------------#
 
 sub new {
-	my ($class, $passwdFile) = @_;
+	my ($proto, $passwdFile) = @_;
+
+	my $class = ref($proto) || $proto;
 	my ($self) = {};
 	bless ($self, $class);
 
 	$self->{'PASSWD'} = $passwdFile;
 	$self->{'ERROR'} = "";
-
+	$self->{'LOCK'} = 0;
+	$self->{'OPEN'} = 0;
+	
 	return $self;
 }
 
@@ -46,8 +82,6 @@ sub htCheckPassword {
 	my ($self) = shift;
 	my ($Id, $pass) = @_;
 
-	my ($passwdFile) = $self->{'PASSWD'};
-
 	my ($cryptPass) = $self->fetchPass($Id);
 
 	if (!$cryptPass) { return undef; }
@@ -57,7 +91,8 @@ sub htCheckPassword {
 	if ($fooCryptPass eq $cryptPass) {
 		return 1;
 	} else {
-		$self->{'ERROR'} = "Old passwords do not match.\n";
+		$self->{'ERROR'} = __PACKAGE__."::htCheckPassword - Passwords do not match.";
+		carp $self->error() unless caller ne $self;
 		return 0;
 	}
 }
@@ -73,63 +108,60 @@ sub htpasswd {
 
 	if (!defined($oldPass)) { $noOld=1;}
 	if (defined($oldPass) && $oldPass =~ /^\d$/) {
-		if ($oldPass == 1) {
+		if ($oldPass) {
 			$newPass = $Id unless $newPass;
 			my ($newEncrypted) = $self->CryptPasswd($newPass);
 			return $self->writePassword($Id, $newEncrypted);
-			exit;
 		}
 	}
 
-# New Entry
-if ($noOld == 1) {
-		my ($passwdFile) = $self->{'PASSWD'};
+	# New Entry
+	if ($noOld) {
+	    my ($passwdFile) = $self->{'PASSWD'};
 
-	# Encrypt new password string
+		# Encrypt new password string
 
-	my ($passwordCrypted) = $self->CryptPasswd($newPass);
+		my ($passwordCrypted) = $self->CryptPasswd($newPass);
 
-	if ($self->fetchPass($Id)) {
+    		$self->_open();
 
-		# User already has a password in the file. 
-		$self->{'ERROR'} = "$Id already exists in $passwdFile";
-		print $self->error() . "\n";
-		return undef;
+		if ($self->fetchPass($Id)) {
+
+			# User already has a password in the file. 
+			$self->{'ERROR'} = __PACKAGE__. "::htpasswd - $Id already exists in $passwdFile";
+			carp $self->error();
+
+			$self->_close();  
+			return undef;
+		} else {
+			# If we can add the user.
 	
+	    		seek(FH, 0, SEEK_END);
+	    		print FH "$Id\:$passwordCrypted\n";
+	    
+	    		$self->_close();  
+	    		return 1;
+		}
+
+    		$self->_close();
+
 	} else {
-		# If we can add the user.
-		if (!open(FH,">>$passwdFile")) {
-			$self->{'ERROR'} = "Cannot append to $passwdFile: $!";
-			print $self->error() . "\n";
+    		$self->_open();
+
+		my ($exists) = $self->htCheckPassword($Id, $oldPass);
+
+		if ($exists) {
+			my ($newCrypted) = $self->CryptPasswd($newPass);
+			return $self->writePassword($Id, $newCrypted);
+		} else {
+			# ERROR returned from htCheckPass
+			$self->{'ERROR'} = __PACKAGE__."::htpasswd - Password not changed.";
+			carp $self->error();
 			return undef;
 		}
 
-		flock(FH, 8);
-		print FH "$Id\:$passwordCrypted\n";
-		flock(FH, 2);
-
-		if (!close(FH)) {
-			$self->{'ERROR'} = "Cannot close $passwdFile: $!";
-			print $self->error() . "\n";
-			return undef;
-		}
-	
-		return 1;
-	}
-	} # end if $noOld == 1
-else {
-	my ($exists) = $self->htCheckPassword($Id, $oldPass);
-
-	if ($exists) {
-		my ($newCrypted) = $self->CryptPasswd($newPass);
-		return $self->writePassword($Id, $newCrypted);
-	} else {
-		# ERROR returned from htCheckPass
-		print $self->error;
-		return undef;
-	}
-	
-}
+    		$self->_close();
+    	}
 } # end htpasswd
 
 #-----------------------------------------------------------#
@@ -143,11 +175,9 @@ sub htDelete {
 	# Loop through the file, building a cache of exising records
 	# which don't match the Id.
 
-	if (!open(FH,"<$passwdFile")) {
-		$self->{'ERROR'} = "cannot open $passwdFile: $!";
-		print $self->error() . "\n";
-		return undef;
-	}
+	$self->_open();
+
+	seek(FH, 0, SEEK_SET);
 	while (<FH>) {
 
 		if (/^$Id\:/) {
@@ -157,40 +187,28 @@ sub htDelete {
 		}
 	}
 
-	if (!close(FH)) {
-		$self->{'ERROR'} = "cannot close $passwdFile: $!";
-		print $self->error() . "\n";
-		return undef;
-	}
 
 	# Write out the @cache if needed.
 
 	if ($return) {
 
-		if (!open(FH,">$passwdFile")) {
-			$self->{'ERROR'} = "cannot open $passwdFile: $!";
-			print $self->error() . "\n";
-			return undef;
-		}
+	    # Return to beginning of file
+	    seek(FH, 0, SEEK_SET);
+	    
+	    while (@cache) { 
+		print FH shift (@cache); 
+	    }
 
-		flock(FH, 2) || print "Can't lock file";
+	    # Cut everything beyond current position
+	    truncate(FH, tell(FH));
 
-		while (@cache) { 
-			print FH shift (@cache); 
-		}
-
-		flock(FH, 8) || print "Can't unlock file";
-
-		if (!close(FH)) {
-			$self->{'ERROR'} = "cannot close $passwdFile: $!";
-			print $self->error() . "\n";
-			return undef;
-		}
 	} else {
-		$self->{'ERROR'} = "user $Id no found in $passwdFile: $!";
-		print $self->error() . "\n";
+	    $self->{'ERROR'} = __PACKAGE__. "::htDelete - User $Id not found in $passwdFile: $!";
+	    carp $self->error();
 	}
 
+	$self->_close();
+
 	return $return;
 }
 
@@ -201,34 +219,22 @@ sub fetchPass {
 	my ($Id) = @_;
 	my ($passwdFile) = $self->{'PASSWD'};
 
-	if (!open(FH,"<$passwdFile")) {
-		$self->{'ERROR'} = "cannot open $passwdFile: $!";
-		print $self->error() . "\n";
-		return undef;
-	}
+	my $passwd = 0;
 
+	$self->_open();
+	
 	while (<FH>) {
 		chop;
-		if ( /^$Id\:/ ) {
-			if (!close(FH)) {
-				$self->{'ERROR'} = "cannot close $passwdFile: $!";
-				print $self->error() . "\n";
-				return undef;
-			}
-
-			$_ =~ s/^[^:]*://;
-			return $_;
+		my @tmp = split(/:/,$_,3);
+		if ( $tmp[0] eq $Id ) {
+		    $passwd = $tmp[1];
+		    last;
 		}
 	}
 
-	# Password not found
-	if (!close(FH)) {
-		$self->{'ERROR'} = "cannot close $passwdFile: $!";
-		print $self->error() . "\n";
-		return undef;
-	}
+	$self->_close();
 
-	return 0;
+	return $passwd;
 }
 
 #-----------------------------------------------------------#
@@ -241,17 +247,93 @@ sub writePassword {
 	my (@cache);
 
 	my ($return);
-
-	if (!open(FH,"<$passwdFile")) {
-		$self->{'ERROR'} = "cannot open $passwdFile: $!";
-		print $self->error() . "\n";
-		return undef;
-	}
+	
+	$self->_open();
+	seek(FH, 0, SEEK_SET);
 
 	while (<FH>) {
 
-		if ( /^$Id\:/ ) {
-			push (@cache, "$Id\:$newPass\n");
+	    my @tmp = split(/:/,$_,3);
+	    if ( $tmp[0] eq $Id ) {
+		my $info = $tmp[2] ? $tmp[2] : "";
+	        push (@cache, "$Id\:$newPass\:$info\n");
+	        $return = 1; 
+
+	    } else {
+	        push (@cache, $_);
+	    }
+	}
+
+	# Write out the @cache, if needed.
+
+	if ($return) {
+	    
+	    # Return to beginning of file
+	    seek(FH, 0, SEEK_SET);
+
+		while (@cache) { 
+			print FH shift (@cache); 
+		}
+
+	    # Cut everything beyond current position
+	    truncate(FH, tell(FH));
+
+	} else {
+		$self->{'ERROR'} = __PACKAGE__. "::writePassword - User $Id not found in $passwdFile: $!";
+		carp $self->error() . "\n";
+	}
+
+	$self->_close();
+
+	return $return;
+}
+
+#-----------------------------------------------------------#
+
+sub fetchInfo {
+	my ($self) = shift;
+	my ($Id) = @_;
+	my ($passwdFile) = $self->{'PASSWD'};
+
+	my $info = 0;
+
+	$self->_open();
+	
+	while (<FH>) {
+		chop;
+		my @tmp = split(/:/,$_,3);
+		if ( $tmp[0] eq $Id ) {
+		    $info = $tmp[2];
+		    last;
+		}
+	}
+
+	$self->_close();
+
+	return $info;
+}
+
+#-----------------------------------------------------------#
+
+sub writeInfo {
+	my ($self) = shift;
+	my ($Id, $newInfo) = @_;
+
+	my ($passwdFile) = $self->{'PASSWD'};
+	my (@cache);
+
+	my ($return);
+	
+	$self->_open();
+	seek(FH, 0, SEEK_SET);
+
+	while (<FH>) {
+
+	        my @tmp = split(/:/,$_,3);
+
+		if ( $tmp[0] eq $Id ) {
+			chomp $tmp[1] if (@tmp == 2); # Cut out EOL if there was no info
+			push (@cache, "$Id\:$tmp[1]\:$newInfo\n");
 			$return = 1; 
 
 		} else {
@@ -259,40 +341,27 @@ sub writePassword {
 		}
 	}
 
-	if (!close(FH)) {
-		$self->{'ERROR'} = "cannot close $passwdFile: $!";
-		print $self->error() . "\n";
-		return undef;
-	}
-
 	# Write out the @cache, if needed.
 
 	if ($return) {
-
-		if (!open(FH, ">$passwdFile")) {
-			$self->{'ERROR'} = "cannot open $passwdFile: $!";
-			print $self->error() . "\n";
-			return undef;
-		}
-
-		flock(FH,2) || croak "Can't lock file";
+	    
+	    # Return to beginning of file
+	    seek(FH, 0, SEEK_SET);
 
 		while (@cache) { 
 			print FH shift (@cache); 
 		}
 
-		flock(FH,8) || croak "Can't lock file";
+	    # Cut everything beyond current position
+	    truncate(FH, tell(FH));
 
-		if (!close(FH)) {
-			$self->{'ERROR'} = "cannot close $passwdFile: $!";
-			print $self->error() . "\n";
-			return undef;
-		}
 	} else {
-		$self->{'ERROR'} = "user $Id not found in $passwdFile: $!";
-		print $self->error() . "\n";
+		$self->{'ERROR'} = __PACKAGE__. "::writeInfo - User $Id not found in $passwdFile: $!";
+		carp $self->error() . "\n";
 	}
 
+	$self->_close();
+
 	return $return;
 }
 
@@ -314,7 +383,75 @@ sub CryptPasswd {
 
 #-----------------------------------------------------------#
 
-sub DESTROY {};
+sub DESTROY { close(FH); };
+
+#-----------------------------------------------------------#
+
+    sub _lock {
+	my ($self) = shift;
+	
+	# Lock if we don't have the lock
+        flock(FH, LOCK_EX) if($self->{'LOCK'} == 0);
+
+	# We have the lock
+	$self->{'LOCK'} = 1;
+
+	# Seek to head
+        seek(FH, 0, SEEK_SET);
+    }
+
+#-----------------------------------------------------------#
+
+    sub _unlock {
+	my ($self) = shift;
+
+        flock(FH, LOCK_UN);
+
+	$self->{'LOCK'} = 0;
+    }
+
+#-----------------------------------------------------------#
+
+sub _open {
+    my ($self) = shift;
+
+    if($self->{'OPEN'} > 0) {
+	$self->{'OPEN'}++;
+	$self->_lock();
+	return;
+    }
+
+    my $passwdFile = $self->{'PASSWD'};
+    if (!open(FH,"+<$passwdFile")) {
+	$self->{'ERROR'} = __PACKAGE__. "::fetchPass - Cannot open $passwdFile: $!";
+	croak $self->error();
+    }
+
+    $self->{'OPEN'}++;
+    $self->_lock();
+}
+
+#-----------------------------------------------------------#
+
+sub _close {
+    my ($self) = shift;
+    $self->_unlock();
+
+    $self->{'OPEN'}--;
+
+    if($self->{'OPEN'} > 0) { return; }
+
+    if (!close(FH)) {
+	my $passwdFile = $self->{'PASSWD'};
+	$self->{'ERROR'} = __PACKAGE__. "::htDelete - Cannot close $passwdFile: $!";
+	carp $self->error();
+	return undef;
+    }
+
+
+}
+
+#-----------------------------------------------------------#
 
 1; 
 
@@ -342,7 +479,7 @@ Apache::Htpasswd - Manage Unix crypt-style password file.
     $foo->htpasswd("zog", "new-password", 1);
         
     # Check that a password is correct
-    $pwdFile->htCheckPass("zog", "password");
+    $pwdFile->htCheckPassword("zog", "password");
 
     # Fetch an encrypted password 
     $foo->fetchPass("foo");
@@ -353,6 +490,12 @@ Apache::Htpasswd - Manage Unix crypt-style password file.
     # If something fails, check error
     $foo->error;
 
+    # Write in the extra info field
+    $foo->writeInfo("login", "info");
+
+    # Get extra info field for a user
+    $foo->fetchInfo("login");
+
 =head1 DESCRIPTION
 
 This module comes with a set of methods to use with htaccess password
@@ -360,7 +503,8 @@ files. These files (and htaccess) are used to do Basic Authentication
 on a web server.
 
 The passwords file is a flat-file with login name and their associated
-crypted password.
+crypted password. You can use this for non-Apache files if you wish, but
+it was written specifically for .htaccess style files.
 
 =head2 FUNCTIONS
 
@@ -424,6 +568,20 @@ Returns 0 if login is invalid.
 Returns undef otherwise.
 
 
+=item fetchInfo("login");
+
+Returns additional information if succeeds.
+Returns 0 if login is invalid.
+Returns undef otherwise.
+
+
+=item writeInfo("login", "info");
+
+Will replace the additional information for the login.
+Returns 0 if login is invalid.
+Returns undef otherwise.
+
+
 =item CryptPasswd("password", "salt");
 
 Will return an encrypted password using 'crypt'. If I<salt> is
@@ -439,6 +597,7 @@ by running these commands:
 
    perl Makefile.PL
    make
+   make test
    make install
    make clean
 
@@ -460,6 +619,34 @@ The latest version of Apache::Htpasswd should always be available from:
 Visit <URL:http://www.perl.com/CPAN/> to find a CPAN
 site near you.
 
+=head1 VERSION
+
+$Revision: 1.3 $ $Date: 2000/04/04 15:00:13 $
+
+=head1 CHANGES
+
+$Log: Htpasswd.pm,v $
+
+Revision 1.5.1  2001/03/26 08:25:38 kevin
+Fixed another newline problem
+
+Revision 1.5  2001/03/15 01:50:12 kevin
+Fixed bug to remove newlines
+
+Revision 1.4  2001/02/23 08:23:46 kevin
+Added support for extra info fields
+
+Revision 1.3  2000/04/04 15:00:15 meltzek
+Made file locking safer to avoid race conditions. Fixed
+typo in docs.
+
+Revision 1.2  1999/01/28 22:43:45  meltzek
+Added slightly more verbose error croaks. Made sure error from htCheckPassword is only called when called directly, and not by $self.
+
+Revision 1.1  1998/10/22 03:12:08  meltzek
+Slightly changed how files lock.
+Made more use out of carp and croak.
+Made sure there were no ^M's as per Randal Schwartz's request.
 
 
 =head1 BUGS
@@ -468,16 +655,22 @@ None knows at time of writting.
 
 =head1 AUTHOR INFORMATION
 
-Copyright 1998, Kevin Meltzer.  All rights reserved.  It may
+Copyright 1998..2001, Kevin Meltzer.  All rights reserved.  It may
 be used and modified freely, but I do request that this copyright
 notice remain attached to the file.  You may modify this module as you
 wish, but if you redistribute a modified version, please attach a note
 listing the modifications you have made.
 
+This is released under the same terms as Perl itself.
+
 Address bug reports and comments to:
-kmeltz@cris.com
+perlguy@perlguy.com
 
 The author makes no warranties, promises, or gaurentees of this software. As with all
 software, use at your own risk.
 
+=head1 SEE ALSO
+
+L<Apache::Htgroup>
+
 =cut
